from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.tokens import TokenEncoder


@dataclass
class Nested:
    something: str


@dataclass
class Message1(Message):
    something: str
    nested: Nested


class TestTokenEncoder:

    def test__should_encode_and_decode_correctly(self):

        # -- Arrange ---------------------------------------------------------

        obj = Message1(
            something='something',
            nested=Nested(something='something nested'),
        )

        uut = TokenEncoder(
            schema=Message1,
            secret='123',
        )

        # -- Act -------------------------------------------------------------

        encoded = uut.encode(obj=obj)
        decoded = uut.decode(encoded_jwt=encoded)

        # -- Assert ----------------------------------------------------------

        assert isinstance(decoded, Message1)
        assert decoded != encoded
        assert decoded == obj



from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.residual_mix import \
    ResidualMixEmissions, ResidualMixTechnologyDistribution


@dataclass
class ResidualMixEmissionsUpdate(Message):
    """
    TODO
    """
    emissions: ResidualMixEmissions


@dataclass
class ResidualMixTechnologyDistributionUpdate(Message):
    """
    TODO
    """
    distributions: ResidualMixTechnologyDistribution



# from typing import List
# from sqlalchemy import orm, asc, desc, and_
#
# from energytt_platform.sql import SqlQuery
# from energytt_platform.models.meteringpoints import MeteringPointType
#
# from .models import (
#     MeteringPointFilters,
#     MeteringPointOrdering,
#     MeteringPointOrderingKeys,
#     DbMeteringPoint,
#     DbMeteringPointTechnology,
#     DbMeteringPointAddress,
#     DbMeteringPointDelegate,
#     DbTechnology,
# )
#
#
# # -- MeteringPoints ----------------------------------------------------------
#
#
# class MeteringPointQuery(SqlQuery):
#     """
#     Query DbMeteringPoint.
#     """
#     def _get_base_query(self) -> orm.Query:
#         return self.session.query(DbMeteringPoint)



import operator


class ArithmeticDict(dict):
    """
    A dictionary that supports arithmetic on all values of the dict.

    TODO Fix type hints etc.
    """

    def __repr__(self):
        return '%s<%s>' % (
            self.__class__.__name__,
            super(ArithmeticDict, self).__repr__(),
        )

    def __add__(self, other):
        """
        :param EmissionValues|int|float other:
        :rtype: EmissionValues
        """
        return self.__do_arithmetic(other, operator.__add__)

    def __radd__(self, other):
        """
        :param EmissionValues|int|float other:
        :rtype: EmissionValues
        """
        return self + other

    def __mul__(self, other):
        """
        :param EmissionValues|int|float other:
        :rtype: EmissionValues
        """
        return self.__do_arithmetic(other, operator.__mul__)

    def __rmul__(self, other):
        """
        :param EmissionValues|int|float other:
        :rtype: EmissionValues
        """
        return self * other

    def __truediv__(self, other):
        """
        :param EmissionValues|int|float other:
        :rtype: EmissionValues
        """
        return self.__do_arithmetic(other, operator.__truediv__)

    def __do_arithmetic(self, other, calc):
        """
        :param EmissionValues|int|float other:
        :param function calc:
        :rtype: EmissionValues
        """
        if isinstance(other, dict):
            keys = set(list(self.keys()) + list(other.keys()))
            return self.__class__(
                **{key: calc(self.get(key) or 0, other.get(key) or 0)
                   for key in keys}
            )
        elif isinstance(other, (int, float)):
            return self.__class__(
                **{key: calc(self.get(key) or 0, other)
                   for key in self.keys()}
            )

        return NotImplemented



from typing import Dict, Any
from urllib.parse import urlencode
import urllib.parse as urlparse


def append_query_parameters(url: str, query_extra: Dict[str, Any]) -> str:
    """
    Provided an URL, this function adds (or overwrites) one or more
    query parameters, and returns the updated URL.

    :param url: The URL to add query parameters to
    :param query_extra: The query parameters to add
    :return: The URL with added query parameters
    """
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(query_extra)
    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Type, TypeVar, Generic, Optional


@dataclass
class Serializable:
    """
    Base class for dataclasses that can be serialized and deserialized.
    Subclasses must be defined as dataclasses.
    """
    pass


TSerializable = TypeVar('TSerializable', bound=Serializable)
TSerialized = TypeVar('TSerialized')


class Serializer(Generic[TSerialized]):
    """
    An interface for serializing and deserializing dataclasses.
    """

    @abstractmethod
    def serialize(
            self, 
            obj: TSerializable,
            schema: Optional[Type[TSerializable]] = None,
    ) -> TSerialized:
        """
        Serialize an object.
        """
        raise NotImplementedError

    @abstractmethod
    def deserialize(
            self,
            data: TSerialized,
            schema: Type[TSerializable],
            validate: bool = True,
    ) -> TSerializable:
        """
        Deserialize an object.
        """
        raise NotImplementedError



"""
conftest.py according to pytest docs:
https://docs.pytest.org/en/2.7.3/plugins.html?highlight=re#conftest-py-plugins
"""
import time
import random
import pytest
from uuid import uuid4
from testcontainers.core.container import DockerContainer

from energytt_platform.models.tech import Technology, TechnologyType
from energytt_platform.bus import (
    MessageBroker,
    Message,
    messages as m,
    get_default_broker,
)


# -- Fixtures ----------------------------------------------------------------


@pytest.fixture(scope='function')
def kafka_container():
    """
    TODO
    """
    kafka_docker = DockerContainer('landoop/fast-data-dev:latest')
    kafka_docker.env.update({'ADV_HOST': 'localhost'})
    kafka_docker.ports.update({
        2181: 2181,
        3030: 3030,
        8081: 8081,
        8082: 8082,
        8083: 8083,
        9581: 9581,
        9582: 9582,
        9583: 9583,
        9584: 9584,
        9585: 9585,
        9092: 9092,
    })

    with kafka_docker as container:
        time.sleep(5)
        yield container


@pytest.fixture(scope='function')
def broker(kafka_container: DockerContainer) -> MessageBroker:
    """
    TODO
    """
    return _create_test_broker(kafka_container)


@pytest.fixture(scope='function')
def broker2(kafka_container: DockerContainer) -> MessageBroker:
    """
    TODO
    """
    return _create_test_broker(kafka_container)


@pytest.fixture(scope='function')
def msg1() -> Message:
    """
    TODO
    """
    return m.TechnologyUpdate(
        technology=Technology(
            tech_code=str(uuid4()),
            fuel_code=str(uuid4()),
            type=random.choice(list(TechnologyType)),
        ),
    )


@pytest.fixture(scope='function')
def msg2() -> Message:
    """
    TODO
    """
    return m.TechnologyUpdate(
        technology=Technology(
            tech_code=str(uuid4()),
            fuel_code=str(uuid4()),
            type=random.choice(list(TechnologyType)),
        ),
    )


@pytest.fixture(scope='function')
def msg3() -> Message:
    """
    TODO
    """
    return m.TechnologyUpdate(
        technology=Technology(
            tech_code=str(uuid4()),
            fuel_code=str(uuid4()),
            type=random.choice(list(TechnologyType)),
        ),
    )


# -- Helpers -----------------------------------------------------------------


def _create_test_broker(kafka_container: DockerContainer) -> MessageBroker:
    """
    Creates a new message broker instance with a unique Consumer Group ID.
    """
    host = kafka_container.get_container_host_ip()
    port = kafka_container.get_exposed_port(9092)
    server = f'{host}:{port}'

    return get_default_broker(
        group=str(uuid4()),
        servers=[server],
    )



import jwt
from typing import Generic, TypeVar, Type

from energytt_platform.serialize import simple_serializer


TToken = TypeVar('TToken')


class TokenEncoder(Generic[TToken]):
    """
    Generic helper-class to encode and decode dataclasses to and from JWT.
    """

    class EncodeError(Exception):
        """
        Raised when encoding fails.
        """
        pass

    class DecodeError(Exception):
        """
        Raised when decoding fails.
        """
        pass

    HS256 = 'HS256'
    RS256 = 'RS256'

    def __init__(self, schema: Type[TToken], secret: str, alg: str = HS256):
        """
        TODO

        :param schema:
        :param secret:
        """
        self.schema = schema
        self.secret = secret
        self.alg = alg

        # TODO Detect schema from Generic type instead of parameter

    def encode(self, obj: TToken) -> str:
        """
        TODO

        :param obj:
        :return:
        """
        payload = simple_serializer.serialize(
            obj=obj,
            schema=self.schema,
        )

        # TODO Raise EncodeError

        return jwt.encode(
            payload=payload,
            key=self.secret,
            algorithm=self.alg,
        )

    def decode(self, encoded_jwt: str) -> TToken:
        """
        TODO

        :param encoded_jwt:
        :return:
        """
        try:
            payload = jwt.decode(
                jwt=encoded_jwt,
                key=self.secret,
                algorithms=[self.alg],
            )
        except jwt.DecodeError as e:
            raise self.DecodeError(str(e))

        return simple_serializer.deserialize(
            data=payload,
            schema=self.schema,
        )



import pytest
from sqlalchemy import orm, exc

from energytt_platform.sql import SqlQuery

from .db import db
from .models import DbTestModel


# -- Helpers -----------------------------------------------------------------


class TestModelQuery(SqlQuery):
    """
    Query DbTestModel.
    """
    def _get_base_query(self) -> orm.Query:
        return self.session.query(DbTestModel)


# -- Fixtures ----------------------------------------------------------------


@pytest.fixture(scope='function')
def seeded_session(session: db.Session):
    """
    TODO
    """
    session.begin()

    try:
        session.add(DbTestModel(string_field='s1', integer_field=1))
        session.add(DbTestModel(string_field='s1', integer_field=2))
        session.add(DbTestModel(string_field='s2', integer_field=1))
        session.add(DbTestModel(string_field='s2', integer_field=2))
    except:
        session.rollback()
    else:
        session.commit()

    yield session


# -- Tests -------------------------------------------------------------------


class TestQueries:
    """
    TODO
    """

    # -- filter() ------------------------------------------------------------

    def test__filter__no_results_exists__should_apply_filter_and_return_nothing(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == 'FOO-BAR')

        count = query.count()
        results = query.all()

        # -- Assert ----------------------------------------------------------

        assert count == 0
        assert len(results) == 0

    @pytest.mark.parametrize('value', ['s1', 's2'])
    def test__filter__results_exists__should_apply_filter_and_return_correct_results(
            self,
            value: str,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == value)

        count = query.count()
        results = query.all()

        # -- Assert ----------------------------------------------------------

        assert count == 2
        assert len(results) == 2
        assert all(result.string_field == value for result in results)

    # -- filter_by() ---------------------------------------------------------

    def test__filter_by__no_results_exists__should_apply_filter_and_return_nothing(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter_by(string_field='FOO-BAR')

        count = query.count()
        results = query.all()

        # -- Assert ----------------------------------------------------------

        assert count == 0
        assert len(results) == 0

    @pytest.mark.parametrize('value', ['s1', 's2'])
    def test__filter_by__results_exists__should_apply_filter_and_return_correct_results(
            self,
            value: str,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter_by(string_field=value)

        count = query.count()
        results = query.all()

        # -- Assert ----------------------------------------------------------

        assert count == 2
        assert len(results) == 2
        assert all(result.string_field == value for result in results)

    # -- one() ---------------------------------------------------------------

    def test__one__one_result__should_return_correct_result(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == 's1') \
            .filter(DbTestModel.integer_field == 1)

        count = query.count()
        result = query.one()

        # -- Assert ----------------------------------------------------------

        assert count == 1
        assert result.string_field == 's1'
        assert result.integer_field == 1

    def test__one__no_results__should_raise_no_result_error(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act + Assert ----------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == 's1') \
            .filter(DbTestModel.integer_field == 9999)

        with pytest.raises(exc.NoResultFound):
            query.one()

        assert query.count() == 0

    # -- one_or_none() -------------------------------------------------------

    def test__one_or_none__one_result__should_return_correct_result(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == 's1') \
            .filter(DbTestModel.integer_field == 1)

        count = query.count()
        result = query.one_or_none()

        # -- Assert ----------------------------------------------------------

        assert count == 1
        assert result.string_field == 's1'
        assert result.integer_field == 1

    def test__one_or_none__no_results__should_return_none(
            self,
            seeded_session: db.Session,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        query = TestModelQuery(seeded_session) \
            .filter(DbTestModel.string_field == 's1') \
            .filter(DbTestModel.integer_field == 9999)

        count = query.count()
        result = query.one_or_none()

        # -- Assert ----------------------------------------------------------

        assert count == 0
        assert result is None

    # -- get() ---------------------------------------------------------------

    # def test__get__one_result__should_return_correct_field(
    #         self,
    #         seeded_session: db.Session,
    # ):
    #     """
    #     TODO
    #     """
    #
    #     # -- Act -------------------------------------------------------------
    #
    #     q = TestModelQuery(seeded_session) \
    #         .filter(DbTestModel.string_field == 's1') \
    #         .filter(DbTestModel.integer_field == 1) \
    #         .only(DbTestModel.string_field)
    #
    #     result = TestModelQuery(seeded_session) \
    #         .filter(DbTestModel.string_field == 's1') \
    #         .filter(DbTestModel.integer_field == 1) \
    #         .get(DbTestModel.string_field)
    #
    #     # -- Assert ----------------------------------------------------------
    #
    #     assert result.string_field == 's1'

    # def test__get__no_results__should_return_none(
    #         self,
    #         seeded_session: db.Session,
    # ):
    #     """
    #     TODO
    #     """
    #
    #     # -- Act + Assert ----------------------------------------------------
    #
    #     query = TestModelQuery(seeded_session) \
    #         .filter(DbTestModel.string_field == 's1') \
    #         .filter(DbTestModel.integer_field == 9999)
    #
    #     count = query.count()
    #     result = query.one_or_none()
    #
    #     assert count == 0
    #     assert result is None



from enum import Enum
from typing import Dict, Union

from energytt_platform.arithmetic import ArithmeticDict


TEmissionValue = Union[int, float]


class EmissionLabel(Enum):
    """
    System-wide labels of known emission types.
    """
    CO2 = 'CO2'
    SO2 = 'SO2'


class EmissionValues(ArithmeticDict, Dict[EmissionLabel, TEmissionValue]):
    """
    Represents a set of emissions indexed by label.

    Units of values depends on the context of their use (gram, gram/Wh, etc).
    """
    pass



from typing import Optional
from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.delegates import MeteringPointDelegate


@dataclass
class MeteringPointOwnerUpdate(Message):
    """
    TODO
    TODO Received by auth-service to determine the owner, who can grant delegates on this GSRN number
    """
    gsrn: str
    subject: Optional[str]


@dataclass
class MeteringPointDelegateGranted(Message):
    """
    An actor (identified by its subject) has been delegated
    access to a MeteringPoint.
    """
    delegate: MeteringPointDelegate


@dataclass
class MeteringPointDelegateRevoked(Message):
    """
    An actor (identified by its subject) has had its delegated
    access to a MeteringPoint revoked.
    """
    delegate: MeteringPointDelegate



import sqlalchemy as sa

from .db import db


class DbTestModel(db.ModelBase):
    __tablename__ = 'test'

    id = sa.Column(sa.Integer(), primary_key=True, index=True)
    string_field = sa.Column(sa.String(), index=True, nullable=False)
    integer_field = sa.Column(sa.Integer(), index=True, nullable=False)



"""
Global topics on the message bus.
"""

AUTH = 'ETT-AUTH'
CERTS = 'ETT-CERTS'
DELEGATES = 'ETT-DELEGATES'
KEYS = 'ETT-KEYS'
MEASUREMENTS = 'ETT-MEASUREMENTS'
MEASUREMENTS_COMMANDS = 'ETT-MEASUREMENTS-COMMANDS'
METERINGPOINTS = 'ETT-METERINGPOINTS'
METERINGPOINTS_COMMANDS = 'ETT-METERINGPOINTS_COMMANDS'
TECHNOLOGIES = 'ETT-TECHNOLOGIES'
USERS = 'ETT-USERS'



import logging
from flask import Flask
from flask.testing import FlaskClient
from functools import cached_property
from typing import List, Iterable, Tuple, Any, Optional

from .guards import EndpointGuard
from .endpoint import Endpoint
from .endpoints import HealthCheck, OpenApiSchema
from .orchestration import \
    RequestOrchestrator, JsonBodyProvider, QueryStringProvider


class Application(object):
    """
    TODO
    """
    def __init__(self, name: str, secret: str):
        self.name = name
        self.secret = secret

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._flask_app(*args, **kwargs)

    @classmethod
    def create(
            cls,
            *args,
            endpoints: Iterable[Tuple[str, str, Endpoint]] = (),
            health_check_path: Optional[str] = None,
            **kwargs,
    ) -> 'Application':
        """
        Create a new instance of an Application
        """

        app = cls(*args, **kwargs)

        # Add endpoints
        for e in endpoints:
            assert 3 <= len(e) <= 4

            method, path, endpoint = e[:3]

            if len(e) == 4:
                guards = e[3]
            else:
                guards = []

            app.add_endpoint(
                method=method,
                path=path,
                endpoint=endpoint,
                guards=guards,
            )

        # Add health check endpoint
        if health_check_path:
            app.add_endpoint(
                method='GET',
                path=health_check_path,
                endpoint=HealthCheck(),
            )

        return app

    @cached_property
    def _flask_app(self) -> Flask:
        """
        TODO
        """
        return Flask(self.name)

    @property
    def wsgi_app(self) -> Flask:
        """
        TODO
        """
        return self._flask_app

    @property
    def test_client(self) -> FlaskClient:
        """
        TODO
        """
        return self._flask_app.test_client()

    def add_endpoint(
            self,
            method: str,
            path: str,
            endpoint: Endpoint,
            guards: List[EndpointGuard] = None,
    ):
        """
        TODO
        """
        if method == 'GET':
            data_provider = QueryStringProvider()
        elif method == 'POST':
            data_provider = JsonBodyProvider()
        else:
            raise RuntimeError('Unsupported HTTP method for endpoints: %s' % method)

        self._flask_app.add_url_rule(
            rule=path,
            endpoint=path,
            methods=[method],
            view_func=RequestOrchestrator(
                endpoint=endpoint,
                data=data_provider,
                secret=self.secret,
                guards=guards,
            ),
        )

    def run_debug(self, host: str, port: int):
        """
        TODO
        """
        self._flask_app.logger.setLevel(logging.DEBUG)
        self._flask_app.run(
            host=host,
            port=port,
            debug=True,
        )



from dataclasses import is_dataclass

import flask
import serpyco
import rapidjson
from abc import abstractmethod
from functools import cached_property
from typing import List, Dict, Any, Optional, Type

from energytt_platform.tokens import TokenEncoder
from energytt_platform.serialize import json_serializer, simple_serializer
from energytt_platform.models.auth import InternalToken

from .context import Context
from .flask import FlaskContext
from .endpoint import Endpoint
from .guards import EndpointGuard, bouncer
from .responses import HttpResponse, BadRequest


# -- Request data ------------------------------------------------------------


class RequestDataProvider(object):
    """
    Provides request parameters.
    """
    @abstractmethod
    def get(self) -> Optional[Dict[str, Any]]:
        """
        TODO
        """
        raise NotImplementedError


class JsonBodyProvider(RequestDataProvider):
    """
    Reads request data from request body JSON.
    """
    def get(self) -> Optional[Dict[str, Any]]:
        """
        TODO
        """
        if not flask.request.data:
            return None

        try:
            return rapidjson.loads(flask.request.data.decode('utf8'))
        except rapidjson.JSONDecodeError:
            raise BadRequest('Invalid JSON body provided')


class QueryStringProvider(RequestDataProvider):
    """
    Reads request data from query parameters.
    """
    def get(self) -> Optional[Dict[str, Any]]:
        """
        TODO
        """
        return dict(flask.request.args)


# -- Orchestrator ------------------------------------------------------------


class RequestOrchestrator(object):
    """
    Orchestrates handling of HTTP requests on behalf of an endpoint.

    Behaves as a Flask endpoint, ie. it is callable without taking any
    parameters, so it plugs into Flask seamlessly.
    """
    def __init__(
            self,
            endpoint: Endpoint,
            data: RequestDataProvider,
            secret: str,
            guards: List[EndpointGuard] = None,
    ):
        self.endpoint = endpoint
        self.data = data
        self.secret = secret
        self.guards = guards

    def __call__(self) -> flask.Response:
        """
        Invoked by Flask to handle a HTTP request.
        """
        try:
            return self._invoke_endpoint()
        except HttpResponse as e:
            return self._handle_http_error(e)
        except Exception as e:
            raise
            return self._handle_exception(e)

    @cached_property
    def _internal_token_encoder(self) -> TokenEncoder[InternalToken]:
        """
        TODO
        """
        return TokenEncoder(
            schema=InternalToken,
            secret=self.secret,
        )

    def _build_context(self) -> Context:
        """
        Creates a new request context.
        """
        return FlaskContext(
            token_encoder=self._internal_token_encoder,
        )

    def _invoke_endpoint(self) -> flask.Response:
        """
        TODO
        """
        context = self._build_context()

        if self.guards:
            bouncer.validate(context, self.guards)

        # -- Define arguments for handler ------------------------------------

        handler_kwargs = {}

        if self.endpoint.requires_context:
            handler_kwargs['context'] = context

        # Deserialize request data (if necessary)
        if self.endpoint.should_parse_request_data:
            # Defaulting to an empty dictionary makes it possible to omit
            # request data for models where all fields are optional
            handler_kwargs['request'] = self._parse_request_data(
                data=self.data.get() or {},
                schema=self.endpoint.request_schema,
            )

        # -- Invoke endpoint -------------------------------------------------

        return_value = self.endpoint.handle_request(**handler_kwargs)

        # -- Parse return value ----------------------------------------------

        if isinstance(return_value, HttpResponse):
            response = return_value
        elif is_dataclass(return_value):
            response = HttpResponse(status=200, model=return_value)
        elif isinstance(return_value, (str, bytes)):
            response = HttpResponse(status=200, body=return_value)
        elif isinstance(return_value, dict):
            response = HttpResponse(status=200, json=return_value)
        elif return_value is None:
            response = HttpResponse(status=200)
        else:
            # TODO Handle this
            raise RuntimeError('ENDPOINT RETURNED INVALID RESPONSE?!?!?')

        # -- Create Flask response -------------------------------------------

        flask_response = flask.Response(
            status=response.status,
            mimetype=response.actual_mimetype,
            headers=response.actual_headers,
            response=response.actual_body,
        )

        for cookie in response.cookies:
            flask_response.set_cookie(
                key=cookie.name,
                value=cookie.value,
                expires=cookie.expires,
                path=cookie.path,
                domain=cookie.domain,
                secure=cookie.secure,
                httponly=cookie.http_only,
                samesite='Strict' if cookie.same_site else None,
            )

        return flask_response

    def _handle_http_error(self, e: HttpResponse) -> flask.Response:
        """
        TODO
        """
        return flask.Response(
            status=e.status,
            response=e.body,
            mimetype='text/html',  # TODO
            headers=e.headers,
        )

    def _handle_exception(self, e: Exception) -> flask.Response:
        """
        TODO
        """
        return flask.Response(
            status=500,
            response='Internal Server Error',
            mimetype='text/html',
        )

    def _parse_request_data(
            self,
            data: Dict[str, Any],
            schema: Type[Any],
    ) -> Any:
        """
        TODO
        """
        try:
            return simple_serializer.deserialize(
                data=data,
                schema=schema,
            )
        except serpyco.exception.ValidationError as e:
            # JSON schema validation failed for request data
            # TODO Parse ValidationError to something useful
            # TODO Format body properly
            raise BadRequest(body=str(e))

    # def _parse_response_object(self, response: Any) -> str:
    #     """
    #     TODO
    #     """
    #     return self.endpoint.response_serializer.dump_json(response)



import rapidjson
from datetime import datetime
from functools import partial, cached_property
from http.cookies import SimpleCookie
from typing import List, Dict, Any, Optional, Union, TypeVar, Generic, Tuple
from dataclasses import dataclass, field, is_dataclass

from energytt_platform.serialize import json_serializer

from .cookies import Cookie


TResponseModel = TypeVar('TResponseModel')


@dataclass
class HttpResponse(Generic[TResponseModel], Exception):
    """
    TODO
    """

    # HTTP Status Code
    status: int

    # Response body
    body: Optional[Union[str, bytes]] = \
        field(default=None)

    # Response body JSON
    json: Optional[Dict[str, Any]] = \
        field(default=None)

    # Response model
    model: Optional[TResponseModel] = \
        field(default=None)

    # Response headers
    headers: Dict[str, str] = \
        field(default_factory=dict)

    # Response cookies
    cookies: Union[List[Cookie], Tuple[Cookie, ...]] = \
        field(default_factory=tuple)

    @cached_property
    def actual_headers(self) -> Dict[str, str]:
        """
        TODO
        """
        headers = {}
        headers.update(self.headers)

        return headers

    @cached_property
    def actual_body(self) -> Optional[Union[str, bytes]]:
        """
        TODO
        """
        if self.body is not None:
            return self.body
        elif self.json is not None:
            return rapidjson.dumps(self.json)
        elif self.model is not None:
            return json_serializer.serialize(self.model)
        else:
            return None

    @cached_property
    def actual_mimetype(self) -> Optional[Union[str, bytes]]:
        """
        TODO
        """
        if self.body is not None:
            return 'text/html'
        elif self.json is not None:
            return 'application/json'
        elif self.model is not None:
            return 'application/json'
        else:
            return 'text/html'


class HttpError(HttpResponse):
    """
    TODO
    """
    def __init__(self, msg: str, status: int, **kwargs):
        kwargs.setdefault('body', f'{status} {msg}')
        super(HttpError, self).__init__(status=status, **kwargs)


class MovedPermanently(HttpResponse):
    """
    HTTP 301 Moved Permanently.
    """
    def __init__(self, url, **kwargs):
        super(MovedPermanently, self).__init__(
            status=301, headers={'Location': url}, **kwargs)


class TemporaryRedirect(HttpResponse):
    """
    HTTP 307 Temporary Redirect.
    """
    def __init__(self, url, **kwargs):
        super(TemporaryRedirect, self).__init__(
            status=307, headers={'Location': url}, **kwargs)


class BadRequest(HttpError):
    """
    HTTP 400 Bad Request.

    Returned by the API in case the client invokes an endpoint but
    validation of input data (either query parameters or POST body) fails.

    This response should be accompanied by the validation errors.
    """
    def __init__(self, **kwargs):
        super(BadRequest, self).__init__(
            status=400, msg='Bad Request', **kwargs)


class Unauthorized(HttpError):
    """
    HTTP 401 Unauthorized.

    Returned by the API in case the client provides an invalid token,
    ie. a token that is expired, belongs to a user that doesn't exist etc.

    This is an indication that the client must acquire a new token.
    """
    def __init__(self, msg: str = 'Unauthorized', **kwargs):
        super(Unauthorized, self).__init__(
            status=401, msg=msg, **kwargs)


class Forbidden(HttpError):
    """
    HTTP 403 Forbidden.

    Returned by the API in case the client provides a token without
    the necessary scope(s).
    """
    def __init__(self, msg: str = 'Forbidden', **kwargs):
        super(Forbidden, self).__init__(
            status=403, msg=msg, **kwargs)


class InternalServerError(HttpError):
    """
    HTTP 500 Internal Server Error.
    """
    def __init__(self, msg: str = 'Internal Server Error', **kwargs):
        super(InternalServerError, self).__init__(
            status=500, msg=msg, **kwargs)



from .serializer import Serializer, Serializable
from .serpyco import SerpycoSimpleSerializer, SerpycoJsonSerializer


simple_serializer = SerpycoSimpleSerializer()

json_serializer = SerpycoJsonSerializer()



from typing import Optional
from dataclasses import dataclass, field

from energytt_platform.serialize import Serializable
from energytt_platform.models.tech import Technology

from .common import EnergyDirection, Address


MeteringPointType = EnergyDirection


@dataclass
class MeteringPoint(Serializable):
    """
    TODO
    """
    gsrn: str
    type: Optional[MeteringPointType] = field(default=None)
    sector: Optional[str] = field(default=None)
    technology: Optional[Technology] = field(default=None)
    address: Optional[Address] = field(default=None)



import pytest

from energytt_platform.api import (
    BadRequest,
    MovedPermanently,
    TemporaryRedirect,
)

from .endpoints import (
    EndpointReturnsGeneric,
    EndpointRaisesGeneric,
    EndpointReturnsResponseModel,
    EndpointWithRequestAndResponseModels,
)


class TestEndpointResponse:
    """
    Testing different responses data types.
    """

    def test__health_check__should_return_status_200(self, client):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        r = client.get('/health')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200

    @pytest.mark.parametrize('obj, response_body', [
        ('string-body', 'string-body'),
        (b'bytes-body', 'bytes-body'),
        # TODO ByteStream, more?
    ])
    def test__endpoint_returns_string_alike__should_return_string_as_body(self, obj, response_body, app, client):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointReturnsGeneric(obj),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.get_data(as_text=True) == response_body
        assert r.headers['Content-Type'] == 'text/html; charset=utf-8'

    def test__endpoint_returns_dict__should_return_dict_as_body(self, app, client):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        response_data = {
            'foo': 'bar',
            'something': {
                'foo': 1,
                'bar': 22.2,
            },
        }

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointReturnsGeneric(response_data),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.json == response_data
        assert r.headers['Content-Type'] == 'application/json'

    @pytest.mark.parametrize('status_code, obj', [
        (301, MovedPermanently('http://something.com/')),
        (307, TemporaryRedirect('http://something.com/')),
        (400, BadRequest()),
    ])
    def test__endpoint_raises_http_response__should_format_response_appropriately(self, status_code, obj, app, client):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRaisesGeneric(obj),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == status_code

    def test__endpoint_raises_exception__should_return_status_500(self, app, client):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRaisesGeneric(Exception('foo bar')),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 500

    def test__endpoint_returns_response_model__should_format_response_appropriately(self, app, client):

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointReturnsResponseModel(),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.headers['Content-Type'] == 'application/json'
        assert r.json == {
            'success': True,
            'something': 'something',
        }

    def test__endpoint_with_request_and_response_models(self, app, client):

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointWithRequestAndResponseModels(),
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            json={'something': 'Hello world'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.headers['Content-Type'] == 'application/json'
        assert r.json == {
            'success': True,
            'something': 'Hello world',
        }


class TestEndpointRedirect:

    @pytest.mark.parametrize('status_code, response', [
        (301, MovedPermanently('http://something.com/')),
        (307, TemporaryRedirect('http://something.com/')),
    ])
    def test__endpoint_returns_redirect(self, status_code, response, app, client):

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='GET',
            path='/something',
            endpoint=EndpointReturnsGeneric(response),
        )

        # -- Act -------------------------------------------------------------

        r = client.get('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == status_code
        assert r.headers['Location'] == 'http://something.com/'

    @pytest.mark.parametrize('status_code, response', [
        (301, MovedPermanently('http://something.com/')),
        (307, TemporaryRedirect('http://something.com/')),
    ])
    def test__endpoint_raises_redirect(self, status_code, response, app, client):

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRaisesGeneric(response),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == status_code
        assert r.headers['Location'] == 'http://something.com/'



from .app import Application
from .cookies import Cookie
from .context import Context
from .endpoint import Endpoint
from .guards import EndpointGuard, ScopedGuard, TokenGuard
from .responses import (
    HttpResponse,
    HttpError,
    MovedPermanently,
    TemporaryRedirect,
    BadRequest,
    Unauthorized,
    Forbidden,
    InternalServerError,
)



from dataclasses import dataclass
from typing import Any

from energytt_platform.api import Endpoint, Context
from energytt_platform.models.auth import InternalToken


class EmptyEndpoint(Endpoint):
    """
    Empty endpoints should always return status 200 and empty body.
    """
    def handle_request(self):
        pass


class EndpointReturnsGeneric(Endpoint):
    """
    Generic endpoint that returns whatever is passed to it's constructor.
    """
    def __init__(self, response: Any):
        self.response = response

    def handle_request(self) -> Any:
        return self.response


class EndpointRaisesGeneric(Endpoint):
    """
    Generic endpoint that raises whatever is passed to it's constructor.
    """
    def __init__(self, response: Any):
        self.response = response

    def handle_request(self) -> Any:
        raise self.response


class EndpointRequiresRequestModel(Endpoint):
    """
    Endpoint that requires an instance of a request model.

    Expected behaviour:
        - The model should be formatted as JSON body
        - "Content-Type" header should be 'application/json'
    """

    @dataclass
    class Request:
        something: str

    def handle_request(self, request: Request):
        return self.Response(
            success=True,
            something='something',
        )


class EndpointReturnsResponseModel(Endpoint):
    """
    Endpoint that returns body as an instance of a response model.

    Expected behaviour:
        - The model should be formatted as JSON body
        - "Content-Type" header should be 'application/json'
    """

    @dataclass
    class Response:
        success: bool
        something: str

    def handle_request(self) -> Response:
        return self.Response(
            success=True,
            something='something',
        )


class EndpointWithRequestAndResponseModels(Endpoint):
    """
    Endpoint that takes a request model, and returns body
    as an instance of a response model.

    Expected behaviour:
        - The model should be formatted as JSON body
        - Content-Type header should be 'application/json'
    """

    @dataclass
    class Request:
        something: str

    @dataclass
    class Response:
        success: bool
        something: str

    def handle_request(self, request: Request) -> Response:
        return self.Response(
            success=True,
            something=request.something,
        )


class EndpointRequiresContextReturnsToken(Endpoint):
    """
    Endpoint that returns context.token as JSON (unmodified).

    Expected behaviour:
        - The token is JSON encoded without mutation
    """

    @dataclass
    class Response:
        token: InternalToken

    def handle_request(self, context: Context) -> Response:
        return self.Response(token=context.token)



from sqlalchemy import orm
from abc import abstractmethod


class SqlQuery(object):
    """
    TODO
    """
    def __init__(self, session: orm.Session, q: orm.Query = None):
        """
        :param session:
        :param q:
        """
        self.session = session
        self.q = q or self._get_base_query()

    @abstractmethod
    def _get_base_query(self) -> orm.Query:
        """
        TODO Describe with example
        """
        raise NotImplementedError

    def __iter__(self):
        return iter(self.q)

    def __getattr__(self, name):
        return getattr(self.q, name)

    def filter(self, *filters):
        """
        TODO Describe with example

        Example usage::

            filter(Model.age==35, Model.country=='Denmark')

        :param filters:
        :return:
        """
        return self.__class__(self.session, self.q.filter(*filters))

    def filter_by(self, **filters):
        """
        TODO Describe with example

        Example usage::

            filter_by(age=35, country='Denmark')

        :param filters:
        :return:
        """
        return self.__class__(self.session, self.q.filter_by(**filters))

    # def unique_join(self, *props, **kwargs):
    #     if props[0] in [c.entity for c in self.q._join_entities]:
    #         return self
    #
    #     return self.__class__(self.session, self.q.join(
    #         *props, **kwargs
    #     ))
    #
    # def lazy_load(self, field):
    #     return self.__class__(self.session, self.q.options(
    #         orm.lazyload(field)
    #     ))
    #
    # def eager_load(self, path, *fields):
    #     if isinstance(path, (list, tuple)):
    #         join = orm.joinedload(*path)
    #     else:
    #         join = orm.joinedload(path)
    #
    #     if fields:
    #         join = join.load_only(*fields)
    #
    #     return self.__class__(self.session, self.q.options(join))
    #
    # def eage_load_path(self, *path):
    #     join = rjoinedload(*path)
    #     if fields:
    #         join = join.load_only(*fields)
    #     return self.__class__(self.session, self.q.options(
    #         join
    #     ))

    def only(self, *fields):
        """
        Narrows down the columns to select.

        TODO Example usage

        :param fields:
        :return:
        """
        return self.__class__(self.session, self.q.options(
            orm.load_only(*fields)
        ))

    def get(self, field):
        """
        Returns value for the field from the first result.

        TODO Example usage

        :param field:
        :return:
        """
        return self.only(field).scalar()

    def exists(self):
        """
        :return: True if result count is >= 1
        """
        return self.count() > 0



from typing import List

from .registry import MessageRegistry
from .kafka import KafkaMessageBroker
from .serialize import MessageSerializer
from .dispatcher import MessageDispatcher
from .broker import MessageBroker, Message


message_registry = MessageRegistry()


def get_default_broker(group: str, servers: List[str]) -> MessageBroker:
    """
    Creates and returns an instance of the default message broker.

    :param group: Consumer group
    :param servers: List of broker servers in the format "IP:PORT"
    :return: An instance of the default message broker
    """
    return KafkaMessageBroker(
        group=group,
        servers=servers,
        serializer=MessageSerializer(registry=message_registry),
    )



from flask.testing import FlaskClient

from energytt_platform.api import Application, HttpResponse

from .endpoints import EndpointReturnsGeneric


class TestHeaders:

    def test__endpoint_returns_redirect(
            self,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO

        :param app:
        :param client:
        :return:
        """

        # -- Arrange ---------------------------------------------------------

        response = HttpResponse(
            status=200,
            headers={
                'Header1': 'Value1',
                'Header2': 'Value2',
            }
        )

        app.add_endpoint(
            method='GET',
            path='/something',
            endpoint=EndpointReturnsGeneric(response),
        )

        # -- Act -------------------------------------------------------------

        r = client.get('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200
        assert r.headers['Header1'] == 'Value1'
        assert r.headers['Header2'] == 'Value2'



from enum import Enum
from dataclasses import dataclass

from energytt_platform.serialize import Serializable


class TechnologyType(Enum):
    """
    System-wide labels of known technologies.
    """
    coal = 'coal'
    nuclear = 'nuclear'
    solar = 'solar'
    wind = 'wind'


@dataclass
class TechnologyCodes(Serializable):
    """
    TODO
    """
    tech_code: str
    fuel_code: str


@dataclass
class Technology(Serializable):
    """
    A technology described by the standard described in the
    EECS Rules Fact Sheet 5: TYPES OF ENERGY INPUTS AND TECHNOLOGIES
    """
    tech_code: str
    fuel_code: str
    type: TechnologyType



from enum import Enum
from serpyco import field
from dataclasses import dataclass
from datetime import date, datetime
from typing import Optional, Generic, TypeVar, List

from energytt_platform.serialize import Serializable


# -- Data objects ------------------------------------------------------------


class EnergyDirection(Enum):
    """
    TODO
    """
    production = 'production'  # E18
    consumption = 'consumption'  # E17


@dataclass
class Address(Serializable):
    """
    TODO Which international standard does this convey to?
    """
    street_code: Optional[str] = field(default=None)
    street_name: Optional[str] = field(default=None)
    building_number: Optional[str] = field(default=None)
    floor_id: Optional[str] = field(default=None)
    room_id: Optional[str] = field(default=None)
    post_code: Optional[str] = field(default=None)
    city_name: Optional[str] = field(default=None)
    city_sub_division_name: Optional[str] = field(default=None)
    municipality_code: Optional[str] = field(default=None)
    location_description: Optional[str] = field(default=None)

    def __str__(self) -> str:
        """
        Creates a string representation of the address.
        """
        return self.format()

    def format(self, separator: str = '\n') -> str:
        """
        Creates a string representation of the address.
        """
        return separator.join(self.parts)

    @property
    def parts(self) -> List[str]:
        """
        TODO
        """
        return []


# -- Value Ranges ------------------------------------------------------------


@dataclass
class DateRange(Serializable):
    """
    A range of dates.
    """
    from_: Optional[date] = field(default=None, dict_key='from')
    to_: Optional[date] = field(default=None, dict_key='to')


@dataclass
class DateTimeRange(Serializable):
    """
    A range of datetimes.
    """
    from_: Optional[datetime] = field(default=None, dict_key='from')
    to_: Optional[datetime] = field(default=None, dict_key='to')


# -- API & Querying ----------------------------------------------------------


TOrderKey = TypeVar('TOrderKey', bound=Enum)


class Order(Enum):
    """
    TODO
    """
    asc = 'asc'
    desc = 'desc'


@dataclass
class ResultOrdering(Serializable, Generic[TOrderKey]):
    """
    Ordering of query results.
    """
    key: Optional[TOrderKey] = field(default=None)
    order: Order = field(default=Order.asc)

    @property
    def asc(self) -> bool:
        return self.order is Order.asc

    @property
    def desc(self) -> bool:
        return self.order is Order.desc



from typing import Dict
from flask import request
from functools import cached_property

from energytt_platform.api.context import Context


class FlaskContext(Context):
    """
    Flask-specific context.
    """

    @cached_property
    def headers(self) -> Dict[str, str]:
        """
        :returns: HTTP request headers
        """
        return dict(request.headers)

    @cached_property
    def cookies(self) -> Dict[str, str]:
        """
        :returns: HTTP request cookies
        """
        return request.cookies



from functools import lru_cache
from typing import Dict, Type, Any, Optional

from .serializer import Serializer, TSerializable

# TODO Describe why:
try:
    import serpyco
except ImportError:
    serpyco = None


@lru_cache
def _get_serializer(schema: Type[TSerializable]) -> serpyco.Serializer:
    """
    TODO
    """
    return serpyco.Serializer(schema, strict=True)


# -- Serializers -------------------------------------------------------------


class SerpycoSimpleSerializer(Serializer[Dict[str, Any]]):
    """
    Serialize and deserialize to and from simple Python types (dictionary).
    """
    def serialize(
            self, obj: TSerializable,
            schema: Optional[Type[TSerializable]] = None,
    ) -> Dict[str, Any]:
        """
        Serializes object to Python.
        """
        if schema is None:
            schema = obj.__class__
        return _get_serializer(schema).dump(obj)

    def deserialize(
            self,
            data: Dict[str, Any],
            schema: Type[TSerializable],
            validate: bool = True,
    ) -> TSerializable:
        """
        Deserialize JSON data to instance of type "cls".
        """
        return _get_serializer(schema) \
            .load(data, validate=validate)


class SerpycoJsonSerializer(Serializer[bytes]):
    """
    Serialize and deserialize to and from JSON (encoded bytes).
    """
    def serialize(
            self,
            obj: TSerializable,
            schema: Optional[Type[TSerializable]] = None,
    ) -> bytes:
        """
        Serializes object to JSON.
        """
        if schema is None:
            schema = obj.__class__
        return _get_serializer(schema).dump_json(obj).encode()

    def deserialize(
            self,
            data: bytes,
            schema: Type[TSerializable],
            validate: bool = True,
    ) -> TSerializable:
        """
        Deserialize JSON data to instance of type "cls".
        """
        return _get_serializer(schema) \
            .load_json(data.decode('utf8'), validate=validate)



"""
conftest.py according to pytest docs:
https://docs.pytest.org/en/2.7.3/plugins.html?highlight=re#conftest-py-plugins
"""
import pytest
from datetime import datetime, timedelta, timezone

from energytt_platform.api import Application
from energytt_platform.models.auth import InternalToken
from energytt_platform.tokens import TokenEncoder


@pytest.fixture(scope='function')
def app(secret: str):
    """
    TODO
    """
    yield Application.create(
        name='Test API',
        secret=secret,
        health_check_path='/health',
    )


@pytest.fixture(scope='function')
def client(app: Application):
    """
    TODO
    """
    yield app.test_client


@pytest.fixture(scope='function')
def secret():
    """
    TODO
    """
    yield 'something secret'


@pytest.fixture(scope='function')
def token_encoder(secret: str):
    """
    TODO
    """
    yield TokenEncoder(
        schema=InternalToken,
        secret=secret,
    )


@pytest.fixture(scope='function')
def valid_token():
    """
    TODO
    """
    yield InternalToken(
        issued=datetime.now(tz=timezone.utc),
        expires=datetime.now(tz=timezone.utc) + timedelta(days=1),
        actor='foo',
        subject='bar',
        scope=['scope1', 'scope2'],
    )


@pytest.fixture(scope='function')
def valid_token_encoded(
        valid_token: InternalToken,
        token_encoder: TokenEncoder[InternalToken],
):
    """
    TODO
    """
    yield token_encoder.encode(valid_token)



from dataclasses import dataclass
from typing import Dict, Any, Optional

from energytt_platform.bus import Message
from energytt_platform.models.common import Address
from energytt_platform.models.tech import TechnologyCodes
from energytt_platform.models.meteringpoints import MeteringPoint


@dataclass
class MeteringPointUpdate(Message):
    """
    Event.

    A MeteringPoint has either been added to the system,
    or an existing MeteringPoint has had its details updated.
    """
    meteringpoint: MeteringPoint


@dataclass
class MeteringPointRemoved(Message):
    """
    Event.

    A MeteringPoint has been remove from the system.
    # TODO Advice to perform Clean-up?
    """
    gsrn: str


@dataclass
class MeteringPointTechnologyUpdate(Message):
    """
    Event.

    Updates technology codes for a MeteringPoint.

    Providing None value for 'codes' indicates that services should
    reset/forget existing technology codes.
    """
    gsrn: str
    codes: Optional[TechnologyCodes]


@dataclass
class MeteringPointAddressUpdate(Message):
    """
    Event.

    Updates address for a MeteringPoint.

    Providing None value for 'address' indicates that services should
    reset/forget existing address.
    """
    gsrn: str
    address: Optional[Address]


@dataclass
class ImportMeteringPoints(Message):
    """
    Command.

    Request system to import a MeteringPoint.

    TODO Describe more
    """
    subject: str
    params: Dict[str, Any]

    # Pre-defined param keys:
    GSRN = 'GSRN'
    WEB_ACCESS_CODE = 'WEB_ACCESS_CODE'
    CVR = 'CVR'
    CPR = 'CPR'



from energytt_platform.bus import message_registry

from .users import (
    UserOnboarded,
)
from .tech import (
    TechnologyUpdate,
    TechnologyRemoved,
)
from .delegates import (
    MeteringPointOwnerUpdate,
    MeteringPointDelegateGranted,
    MeteringPointDelegateRevoked,
)
from .measurements import (
    MeasurementUpdate,
    MeasurementRemoved,
    ImportMeasurements,
)
from .meteringpoints import (
    MeteringPointUpdate,
    MeteringPointRemoved,
    MeteringPointTechnologyUpdate,
    MeteringPointAddressUpdate,
    ImportMeteringPoints,
)


message_registry.add(

    # Users
    UserOnboarded,

    # Delegates
    MeteringPointOwnerUpdate,
    MeteringPointDelegateGranted,
    MeteringPointDelegateRevoked,

    # Technology
    TechnologyUpdate,
    TechnologyRemoved,

    # Measurements
    MeasurementUpdate,
    MeasurementRemoved,
    ImportMeasurements,

    # MeteringPoints
    MeteringPointUpdate,
    MeteringPointRemoved,
    MeteringPointTechnologyUpdate,
    MeteringPointAddressUpdate,

)



from typing import List
from abc import abstractmethod

from .context import Context
from .responses import Unauthorized


class EndpointGuard(object):

    @abstractmethod
    def validate(self, context: Context):
        raise NotImplementedError


# class ServiceGuard(EndpointGuard):
#     """
#     Allows only specific services to access this endpoint.
#     """
#     def __init__(self, *services: Service):
#         self.services = services
#
#
# class IssuerGuard(EndpointGuard):
#     """
#     Allows only specific issuers to access this endpoint.
#     """
#     def __init__(self, *services: Service):
#         self.services = services


class TokenGuard(EndpointGuard):
    """
    Only Allows requests with a valid token provided.
    """
    def validate(self, context: Context):
        if context.token is None:
            raise Unauthorized('Unauthorized')


class ScopedGuard(EndpointGuard):
    """
    Only Allows requests with specific scopes granted.
    """
    def __init__(self, *scopes: str):
        self.scopes = scopes

    def validate(self, context: Context):
        if context.token is None:
            raise Unauthorized('Unauthorized')
        for scope in self.scopes:
            if scope not in context.token.scope:
                raise Unauthorized('Missing scope %s' % scope)  # TODO Write proper message


class Bouncer(object):
    def validate(self, context: Context, guards: List[EndpointGuard]):
        for guard in guards:
            guard.validate(context)


# -- Singletons --------------------------------------------------------------


bouncer = Bouncer()



from wrapt import decorator
from typing import Dict, Any
from sqlalchemy import orm, engine
from functools import cached_property


class SqlEngine(object):
    """
    TODO
    """

    # Shortcut/alias
    Session = orm.Session

    def __init__(self, uri: str, pool_size: int = 1):
        self.uri = uri
        self.pool_size = pool_size

        #
        self._uri = None
        self._engine = None

    @property
    def settings(self) -> Dict[str, Any]:
        """
        TODO
        """
        return {
            'echo': False,
            'pool_pre_ping': True,
            'pool_size': self.pool_size,
        }

    @property
    def engine(self) -> engine.Engine:
        """
        TODO
        """
        if self.uri != self._uri:
            self._uri = self.uri
            self._engine = engine.create_engine(self.uri, **self.settings)
        return self._engine

    @property
    def session_class(self) -> orm.scoped_session:
        """
        TODO
        """
        orm.configure_mappers()
        factory = orm.sessionmaker(bind=self.engine, expire_on_commit=False)
        session_class = orm.scoped_session(factory)
        return session_class

    @cached_property
    def registry(self) -> orm.registry:
        """
        TODO
        """
        return orm.registry()

    @cached_property
    def ModelBase(self) -> orm.declarative_base:
        """
        TODO
        """
        return self.registry.generate_base()

    def make_session(self) -> orm.Session:
        """
        Create a new database session.

        :rtype: sqlalchemy.orm.Session
        """
        return self.session_class()

    def apply_schema(self):
        """
        TODO
        """
        self.ModelBase.metadata.create_all(self.engine)

    def session(self):
        """
        Function decorator which injects a "session" named parameter
        if it doesn't already exists
        """
        @decorator
        def session_decorator(wrapped, instance, args, kwargs):

            session = kwargs.setdefault(
                'session', self.make_session())

            try:
                return wrapped(*args, **kwargs)
            finally:
                session.close()

        return session_decorator

    def atomic(self):
        """
        Function decorator which injects a "session" named parameter
        if it doesn't already exists, and wraps the function in an
        atomic transaction.
        """
        @decorator
        def atomic_wrapper(wrapped, instance, args, kwargs):

            session: orm.Session = kwargs.setdefault(
                'session', self.make_session())

            session.begin()

            try:
                return_value = wrapped(*args, **kwargs)
            except:
                session.rollback()
                raise
            else:
                session.commit()
                return return_value
            finally:
                session.close()

        return atomic_wrapper



from inspect import isclass
from typing import Dict, Union, Type, Optional

from .broker import Message


TDictItem = Union[str, Message, Type[Message]]


class MessageRegistry(Dict[str, Type[Message]]):
    """
    A registry of all messages that the bus knows of.

    Works a a dict where:
        Key = Message type name (str)
        Value = Message class (Serializable)

    TODO Enforce unique names
    """

    @classmethod
    def from_message_types(cls, *types: Type[Message]) -> 'MessageRegistry':
        """
        TODO

        :param types:
        :return:
        """
        return cls({c.__name__: c for c in types})

    def add(self, *message_types: Type[Message]):
        """
        TODO

        :param message_types:
        :return:
        """
        self.update({c.__name__: c for c in message_types})

    def __contains__(self, item: TDictItem) -> bool:
        """
        Check whether an item is known by the registry.

        Item can be either of the following:
            - A string (name of the message type)
            - A class type (the message type itself)
            - An instance of a class (an instance of a message type)
        """
        if isinstance(item, str):
            return item in self.keys()
        elif isclass(item):
            return item in self.values()
        elif isinstance(item, Message):
            return item.__class__ in self.values()
        else:
            return False

    def get(self, item: TDictItem) -> Optional[Message]:
        """
        TODO
        """
        if isinstance(item, str):
            return super(MessageRegistry, self).get(item)
        elif isclass(item):
            return super(MessageRegistry, self).get(item.__name__)
        elif isinstance(item, Message):
            return super(MessageRegistry, self).get(item.__class__.__name__)
        else:
            # TODO:  return something else?
            return None



import pytest
from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.serialize import \
    Serializer, simple_serializer, json_serializer


@dataclass
class Nested:
    something: str


@dataclass
class Message1(Message):
    something: str
    nested: Nested


class TestSerializer:

    @pytest.mark.parametrize('uut', [
        simple_serializer,
        json_serializer,
    ])
    def test__should_serialize_and_deserialize_correctly(self, uut: Serializer):

        # -- Arrange ---------------------------------------------------------

        obj = Message1(
            something='something',
            nested=Nested(something='something nested'),
        )

        # -- Act -------------------------------------------------------------

        serialized = uut.serialize(obj=obj)
        deserialized = uut.deserialize(data=serialized, schema=Message1)

        # -- Assert ----------------------------------------------------------

        assert isinstance(deserialized, Message1)
        assert deserialized == obj



from serpyco import field
from datetime import datetime
from dataclasses import dataclass
from typing import Generic, TypeVar, Union, Dict, Any, Optional

from energytt_platform.serialize import json_serializer, simple_serializer

from .broker import Message
from .registry import MessageRegistry


TWrappedMessage = TypeVar('TWrappedMessage', bound=Message)
TSerializedMessage = Dict[str, Any]


@dataclass
class WrappedMessage(Message, Generic[TWrappedMessage]):
    """
    TODO
    """
    type: str
    time: datetime = field(default_factory=datetime.now)
    msg: Optional[Union[TWrappedMessage, TSerializedMessage]] = field(default=None)


class MessageSerializer(object):
    """
    A serializer specifically for serializing messages to
    and from the event bus.
    """

    class SerializeError(Exception):
        """
        TODO
        """
        pass

    class DeserializeError(Exception):
        """
        TODO
        """
        pass

    def __init__(self, registry: MessageRegistry):
        """
        TODO

        :param registry:
        """
        self.registry = registry

    def serialize(self, msg: Message) -> bytes:
        """
        Wraps message appropriately and JSON serializes it.
        """
        if msg not in self.registry:
            raise self.SerializeError((
                f'Can not serialize of type "{msg.__class__.__name__}": '
                'Type is unknown to the bus.'
            ))

        wrapped_msg = WrappedMessage(
            type=msg.__class__.__name__,
            msg=msg,
        )

        return json_serializer.serialize(
            obj=wrapped_msg,
            schema=WrappedMessage[msg.__class__],
        )

    def deserialize(self, data: bytes) -> Message:
        """
        Deserializes JSON bytestream into a message.
        """
        wrapped_msg = json_serializer.deserialize(
            data=data,
            schema=WrappedMessage[TSerializedMessage],
        )

        if wrapped_msg.type not in self.registry:
            raise self.DeserializeError((
                f'Can not deserialize message of type "{wrapped_msg.type}": '
                'Type is unknown to the bus.'
            ))

        message_cls = self.registry.get(wrapped_msg.type)

        return simple_serializer.deserialize(
            data=wrapped_msg.msg,
            schema=message_cls,
        )



from typing import List, Any, Iterable, Dict
from functools import cached_property
from kafka import KafkaProducer, KafkaConsumer

from energytt_platform.bus.serialize import MessageSerializer
from energytt_platform.bus.broker import MessageBroker, Message, TTopicList


class KafkaMessageBroker(MessageBroker):
    """
    Implementation of Kafka as message bus.
    """
    def __init__(
            self,
            group: str,
            servers: List[str],
            serializer: MessageSerializer,
    ):
        self.group = group
        self.servers = servers
        self.serializer = serializer

    @cached_property
    def _kafka_producer(self) -> KafkaProducer:
        """
        TODO
        """
        return KafkaProducer(
            bootstrap_servers=self.servers,
            value_serializer=self.serializer.serialize,

        )

    @cached_property
    def _kafka_consumer(self) -> KafkaConsumer:
        """
        TODO
        """

        return KafkaConsumer(
            bootstrap_servers=self.servers,
            value_deserializer=self.serializer.deserialize,
            group_id=self.group,
            auto_offset_reset='earliest',
            enable_auto_commit=True,
        )

    def __iter__(self) -> Iterable[Message]:
        """
        Returns an iterable of messages received in any
        of the subscribed topics.
        """
        return (msg.value for msg in self._kafka_consumer)

    def poll(self, timeout: int = 0) -> Dict[str, List[Message]]:
        """
        Polls the broker for at least one message with a timeout.
        Returns messages mapped by topic.

        :param timeout: Timeout in seconds
        """
        res = self._kafka_consumer.poll(timeout_ms=timeout*1000)

        return {
            partition.topic: [record.value for record in record_list]
            for partition, record_list in res.items()
        }

    def poll_list(self, timeout: int = 0) -> List[Message]:
        """
        Polls the broker for at least one message with a timeout.
        Returns a list of messages from any topics subscribed to.

        :param timeout: Timeout in seconds
        """
        res = self._kafka_consumer.poll(timeout_ms=timeout*1000)

        return [
            record.value
            for record_list in res.values()
            for record in record_list
        ]

    def subscribe(self, topics: TTopicList):
        """
        Subscribe to a number of topics.

        :param topics: The topics to subscribe to
        """
        self._kafka_consumer.subscribe(topics)

    def publish(self, topic: str, msg: Any, block=True, timeout=10):
        """
        TODO
        """
        print('PUBLISH: %s' % msg)
        self._kafka_producer.send(topic=topic, value=msg)
        self._kafka_producer.flush()

        # future = self._kafka_producer.send(
        #     topic=topic,
        #     value=msg,
        # )
        #
        # if block:
        #     try:
        #         record_metadata = future.get(timeout=timeout)
        #     except KafkaError as e:
        #         # Decide what to do if produce request failed...
        #         raise self.PublishError(str(e))



from dataclasses import dataclass

from energytt_platform.serialize import Serializable


@dataclass
class MeteringPointDelegate(Serializable):
    """
    An actor (identified by its subject) who has been delegated
    access to a MeteringPoint (identified by its GSRN number).
    """
    subject: str
    gsrn: str

    # TODO Define time period (???)



from typing import List
from dataclasses import dataclass
from datetime import datetime, timezone

from energytt_platform.serialize import Serializable


@dataclass
class InternalToken(Serializable):
    """
    TODO
    """
    issued: datetime
    expires: datetime

    # The user performing action(s)
    actor: str

    # The subject we're working with data on behalf of
    subject: str

    # Scopes granted on subject's data
    # meteringpoints.read, measurements.read, emissions.read, meteringpoint.onboard
    scope: List[str]

    @property
    def is_expired(self) -> bool:
        return self.expires < datetime.now(tz=timezone.utc)



from uuid import uuid4

import pytest
from datetime import datetime
from itertools import product
from flask.testing import FlaskClient
from typing import Dict, Iterable, Any

from energytt_platform.api.testing import CookieTester
from energytt_platform.api import Application, HttpResponse, Cookie

from .endpoints import EndpointReturnsGeneric


def get_cookie_combinations() -> Iterable[Dict[str, Any]]:
    """
    TODO
    """

    combinations = product(
        (None, True, False),     # http_only
        (None, True, False),     # secure
        (None, True, False),     # same_site
        (None, 'domain.com'),    # domain
        (None, '/path'),         # path
        (None, datetime.now()),  # expires
    )

    for http_only, secure, same_site, domain, path, expires in combinations:
        yield {
            'name': str(uuid4()),
            'value': str(uuid4()),
            'http_only': http_only,
            'secure': secure,
            'same_site': same_site,
            'domain': domain,
            'path': path,
            'expires': expires,
        }


class TestCookies:

    @pytest.mark.parametrize('cookie_kwargs', get_cookie_combinations())
    def test__set_one_cookie__should_set_cookie_correctly(
            self,
            cookie_kwargs: Dict[str, Any],
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        response = HttpResponse(
            status=200,
            cookies=(Cookie(**cookie_kwargs),),
        )

        app.add_endpoint(
            method='GET',
            path='/something',
            endpoint=EndpointReturnsGeneric(response),
        )

        # -- Act -------------------------------------------------------------

        r = client.get('/something')

        # -- Assert ----------------------------------------------------------

        CookieTester(r.headers) \
            .assert_has_cookies(cookie_kwargs['name']) \
            .assert_cookie(**cookie_kwargs)

    def test__set_multiple_cookies__should_set_all_cookies_correctly(
            self,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        response = HttpResponse(
            status=200,
            cookies=(
                Cookie(name='Cookie1', value='Value1', same_site=True),
                Cookie(name='Cookie2', value='Value2'),
                Cookie(name='Cookie3', value='Value3'),
            ),
        )

        app.add_endpoint(
            method='GET',
            path='/something',
            endpoint=EndpointReturnsGeneric(response),
        )

        # -- Act -------------------------------------------------------------

        r = client.get('/something')

        # -- Assert ----------------------------------------------------------

        CookieTester(r.headers) \
            .assert_has_cookies('Cookie1', 'Cookie2', 'Cookie3') \
            .assert_cookie(name='Cookie1', value='Value1', same_site=True) \
            .assert_cookie(name='Cookie2', value='Value2') \
            .assert_cookie(name='Cookie3', value='Value3')



from typing import Optional
from datetime import datetime
from wsgiref.headers import Headers
from http.cookies import SimpleCookie


class CookieTester(object):
    """
    Helper for testing cookies in HTTP responses.

    TODO Test this class.
    """
    def __init__(self, headers: Headers):
        """
        :param headers:
        """
        self.cookies = SimpleCookie('\r\n'.join(headers.get_all('Set-Cookie')))

    def assert_has_cookies(self, *names: str) -> 'CookieTester':
        """
        Assert that exactly the specified cookies were set.

        :param names:
        """
        assert tuple(self.cookies.keys()) == names

        return self

    def assert_cookie(
            self,
            name: str,
            value: Optional[str] = None,
            expires: Optional[datetime] = None,
            path: Optional[str] = None,
            comment: Optional[str] = None,
            domain: Optional[str] = None,
            max_age: Optional[str] = None,
            secure: Optional[bool] = None,
            http_only: Optional[bool] = None,
            version: Optional[str] = None,
            same_site: Optional[bool] = None,
    ) -> 'CookieTester':
        """
        Assert content of a cookie.

        :param name:
        :param value:
        :param expires:
        :param path:
        :param comment:
        :param domain:
        :param max_age:
        :param secure:
        :param http_only:
        :param version:
        :param same_site:
        :return:
        """
        if value is not None:
            assert self.cookies[name].value == value
        assert self.cookies[name]['expires'] == \
               (expires.strftime('%a, %d %b %Y %H:%M:%S GMT') if expires is not None else '')
        assert self.cookies[name]['path'] == (path if path is not None else '')
        assert self.cookies[name]['comment'] == (comment if comment is not None else '')
        assert self.cookies[name]['domain'] == (domain if domain is not None else '')
        assert self.cookies[name]['max-age'] == (max_age if max_age is not None else '')
        assert self.cookies[name]['secure'] == (secure if secure else '')
        assert self.cookies[name]['httponly'] == (http_only if http_only else '')
        assert self.cookies[name]['version'] == (version if version is not None else '')
        assert self.cookies[name]['samesite'] == ('Strict' if same_site else '')

        return self

    def get_value(self, name: str) -> Optional[str]:
        """
        Get value of a cookie.
        """
        return self.cookies[name].value



# import json
# import requests
#
# from energytt_platform.serialize import json_serializer
#
# from .models import GetMixEmissionsResponse
#
#
# class ConnectionError(Exception):
#     """
#     Raised when invoking EnergyTypeService results
#     in a connection error
#     """
#     pass
#
#
# class ServiceError(Exception):
#     """
#     Raised when invoking EnergyTypeService results
#     in a status code != 200
#     """
#     def __init__(self, msg, status_code=None, response_body=None):
#         super(ServiceError, self).__init__(msg)
#         self.status_code = status_code
#         self.response_body = response_body
#
#
# # class ServiceUnavailable(Exception):
# #     """
# #     Raised when requesting energy type which is unavailable
# #     for the requested GSRN
# #     """
# #     pass
#
#
# class ServiceClient(object):
#     """
#     TODO
#     """
#     def __init__(self, base_url: str, debug: bool = False):
#         """
#         :param base_url:
#         """
#         self.base_url = base_url
#         self.debug = debug
#
#     def get_url(self, path: str):
#         """
#         :param path:
#         :return:
#         """
#         return f'{self.base_url}{path}'
#
#     def invoke(self, func, path, query, response_schema):
#         """
#         :param Callable func:
#         :param str path:
#         :param dict query:
#         :param Schema response_schema:
#         :rtype obj:
#         """
#         headers = {
#             'Content-type': 'application/json',
#             'accept': 'application/json',
#         }
#
#         url = self.get_url(path)
#
#         try:
#             response = func(
#                 url=url,
#                 params=query,
#                 verify=not self.debug,
#                 headers=headers,
#             )
#         except Exception as e:
#             raise ConnectionError(f'Failed to request request service: {e}')
#
#         if response.status_code != 200:
#             raise ServiceError(
#                 status_code=response.status_code,
#                 response_body=response.content.decode(),
#                 msg=(
#                     f'Request resulted in status code {response.status_code}: '
#                     f'{url}\n\n{response.content}'
#                 ),
#             )
#
#         return json_serializer.deserialize(
#             data=response.content,
#             schema=response_schema,
#         )
#
#         # try:
#         #     response = json_serializer.deserialize(
#         #         data=response.content,
#         #         schema=response_schema,
#         #     )
#         #     # response_json = response.json()
#         #     # response_model = response_schema().load(response_json)
#         # except json.decoder.JSONDecodeError:
#         #     raise ServiceError(
#         #         f'Failed to parse response JSON: {url}\n\n{response.content}',
#         #         status_code=response.status_code,
#         #         response_body=str(response.content),
#         #     )
#         # except marshmallow.ValidationError as e:
#         #     raise ServiceError(
#         #         f'Failed to validate response JSON: {url}\n\n{response.content}\n\n{str(e)}',
#         #         status_code=response.status_code,
#         #         response_body=str(response.content),
#         #     )
#         #
#         # return response
#
#     def get(self, *args, **kwargs):
#         return self.invoke(requests.get, *args, **kwargs)
#
#     def post(self, *args, **kwargs):
#         return self.invoke(requests.post, *args, **kwargs)
#
#     def get_residual_mix(self, sector, begin_from, begin_to) -> GetMixEmissionsResponse:
#         """
#         Returns a dict of emission data for a MeteringPoint.
#         :param list[str] sector:
#         :param datetime.datetime begin_from:
#         :param datetime.datetime begin_to:
#         :rtype: GetMixEmissionsResponse
#         """
#         return self.get(
#             path='/residual-mix',
#             response_schema=GetMixEmissionsResponse,
#             query={
#                 'sector': sector,
#                 'begin_from': begin_from.isoformat(),
#                 'begin_to': begin_to.isoformat(),
#             },
#         )



from typing import Optional
from datetime import datetime
from dataclasses import dataclass, field


@dataclass
class Cookie:
    """
    TODO Test this

    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie
    """
    name: str
    value: str
    http_only: bool = field(default=False)
    secure: bool = field(default=False)
    same_site: bool = field(default=False)
    domain: Optional[str] = field(default=None)
    path: Optional[str] = field(default=None)
    expires: Optional[datetime] = field(default=None)



from energytt_platform.encrypt import aes256_encrypt, aes256_decrypt


def test__aes256__should_encrypt_and_decrypt_correctly():

    # -- Arrange -------------------------------------------------------------

    key = 'This is my encryption key'
    data = 'This is the string i want encrypted'

    # -- Act -----------------------------------------------------------------

    data_encrypted = aes256_encrypt(
        data=data,
        key=key,
    )

    data_decrypted = aes256_decrypt(
        data_encrypted=data_encrypted,
        key=key,
    )

    # -- Assert --------------------------------------------------------------

    assert data_decrypted == data
    assert data_decrypted != data_encrypted



from abc import abstractmethod
from dataclasses import dataclass
from typing import List, Iterable, Callable, Any, Union, Tuple, Dict

from energytt_platform.serialize import Serializable


@dataclass
class Message(Serializable):
    """
    Base-class for messages that can be sent on the bus.
    Inherited classes must remember to use the @dataclass decorator.
    """
    pass


TTopic = str
TTopicList = Union[List[TTopic], Tuple[TTopic, ...]]

TMessageHandler = Callable[[Message], None]


class MessageBroker(object):
    """
    Abstract base-class for publishing and consuming messages
    on the message-bus.
    """

    class PublishError(Exception):
        """
        TODO
        """
        pass

    class DispatchError(Exception):
        """
        TODO
        """
        pass

    @abstractmethod
    def __iter__(self) -> Iterable[Message]:
        """
        Returns an iterable of messages received in any
        of the subscribed topics.
        """
        raise NotImplementedError

    @abstractmethod
    def poll(self, timeout: int = 0) -> Dict[str, List[Message]]:
        """
        Polls the broker for at least one message with a timeout.
        Returns messages mapped by topic.

        :param timeout: Timeout in seconds
        """
        raise NotImplementedError

    @abstractmethod
    def poll_list(self, timeout: int = 0) -> List[Message]:
        """
        Polls the broker for at least one message with a timeout.
        Returns a list of messages from any topics subscribed to.

        :param timeout: Timeout in seconds
        """
        raise NotImplementedError

    @abstractmethod
    def subscribe(self, topics: TTopicList):
        """
        Subscribe to a number of topics.

        :param topics: The topics to subscribe to
        """
        raise NotImplementedError

    def listen(self, topics: TTopicList, handler: TMessageHandler):
        """
        Subscribed to the provided topics and invokes the handler
        with each new message.
        """
        self.subscribe(topics)

        for msg in self:
            handler(msg)

    @abstractmethod
    def publish(self, topic: TTopic, msg: Any, block=False, timeout=10):
        """
        Publish a message to a topic on the bus.

        :param topic: The topic to publish to
        :param msg: The message to publish
        :param block: Whether to block until publishing is complete
        :param timeout: Timeout in seconds (if block=True)
        """
        raise NotImplementedError



from flask.testing import FlaskClient

from energytt_platform.api import Application
from energytt_platform.models.auth import InternalToken

from .endpoints import (
    EndpointRequiresContextReturnsToken,
)


class TestContext:
    """
    TODO
    """

    def test__no_token_provided__should_return_null(
            self,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRequiresContextReturnsToken(),
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200
        assert r.json == {'token': None}

    def test__invalid_token_provided__should_return_null(
            self,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRequiresContextReturnsToken(),
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: INVALID-TOKEN'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200
        assert r.json == {'token': None}

    def test__valid_token_provided__should_return_token_as_json(
            self,
            app: Application,
            client: FlaskClient,
            valid_token: InternalToken,
            valid_token_encoded: str,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EndpointRequiresContextReturnsToken(),
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: {valid_token_encoded}'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200
        assert r.json == {
            'token': {
                'issued': valid_token.issued.isoformat(),
                'expires': valid_token.expires.isoformat(),
                'actor': valid_token.actor,
                'subject': valid_token.subject,
                'scope': valid_token.scope,
            },
        }



from Crypto import Random
from Crypto.Cipher import AES
from hashlib import sha256
from base64 import b64encode, b64decode


_BLOCK_SIZE = 16


def _pad(s: str) -> str:
    return (s + (_BLOCK_SIZE - len(s) % _BLOCK_SIZE)
            * chr(_BLOCK_SIZE - len(s) % _BLOCK_SIZE))


def _unpad(s: bytes) -> bytes:
    return s[:-ord(s[len(s) - 1:])]


def aes256_encrypt(data: str, key: str) -> str:
    """
    AES 256-encrypts a string.

    :param data: The data to encrypt
    :param key: Encryption key
    :returns: Encrypted data (base64 encoded)
    """
    private_key = sha256(key.encode('utf8')).digest()
    data = _pad(data)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return b64encode(
        iv + cipher.encrypt(data.encode('utf8'))).decode('utf8')


def aes256_decrypt(data_encrypted: str, key: str) -> str:
    """
    AES 256-decrypts a string.

    :param data_encrypted: The encrypted data (base64 encoded)
    :param key: Encryption key
    :returns: Decrypted data
    """
    data_encrypted = data_encrypted.encode('utf8')
    private_key = sha256(key.encode('utf8')).digest()
    data_encrypted = b64decode(data_encrypted)
    iv = data_encrypted[:16]
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return _unpad(cipher.decrypt(data_encrypted[16:])).decode('utf8')



from datetime import datetime
from dataclasses import dataclass

from energytt_platform.serialize import Serializable


@dataclass
class GranularCertificate(Serializable):
    """
    A single Granular Certificate.
    """
    id: str
    issued: datetime
    expires: datetime
    begin: datetime
    end: datetime
    sector: str
    amount: int
    technology_code: str
    fuel_code: str



from datetime import datetime
from unittest.mock import Mock

from energytt_platform.bus import messages as m
from energytt_platform.bus.dispatcher import MessageDispatcher


from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.common import DateTimeRange
from energytt_platform.models.measurements import Measurement


@dataclass
class Message1(Message):
    something: str


@dataclass
class Message2(Message):
    something: str


class TestMessageSerializer:

    def test__handler_exists_for_type__should_invoke_handler(self):

        # -- Arrange ---------------------------------------------------------

        msg = Message1(something='something')

        handler1 = Mock()
        handler2 = Mock()

        uut = MessageDispatcher({
            Message1: handler1,
            Message2: handler2,
        })

        # -- Act -------------------------------------------------------------

        uut(msg)

        # -- Assert ----------------------------------------------------------

        handler1.assert_called_once_with(msg)
        handler2.assert_not_called()

    def test__handler_does_not_exist_for_type__should_not_invoke_handler(self):

        # -- Arrange ---------------------------------------------------------

        handler = Mock()

        uut = MessageDispatcher({
            Message1: handler,
        })

        # -- Act -------------------------------------------------------------

        uut(Message2(something='something'))

        # -- Assert ----------------------------------------------------------

        handler.assert_not_called()



from dataclasses import dataclass
from unittest.mock import MagicMock, patch

import pytest

from energytt_platform.bus import Message
from energytt_platform.bus.serialize import MessageSerializer


@dataclass
class Nested:
    something: str


@dataclass
class Message1(Message):
    something: str
    nested: Nested


class TestMessageSerializer:

    def test__should_serialize_and_deserialize_correctly(self):

        # -- Arrange ---------------------------------------------------------

        obj = Message1(
          something='something',
          nested=Nested(something='something nested'),
        )

        registry_mock = MagicMock()
        registry_mock.__contains__.return_value = True
        registry_mock.get.return_value = Message1

        uut = MessageSerializer(registry=registry_mock)

        # -- Act -------------------------------------------------------------

        serialized = uut.serialize(obj)
        deserialized = uut.deserialize(serialized)

        # -- Assert ----------------------------------------------------------

        assert isinstance(deserialized, Message1)
        assert deserialized == obj

    def test__serialize__message_not_in_registry__should_raise_serialize_error(self):

        # -- Arrange ---------------------------------------------------------

        obj = Message1(
          something='something',
          nested=Nested(something='something nested'),
        )

        registry_mock = MagicMock()
        registry_mock.__contains__.return_value = False

        uut = MessageSerializer(registry=registry_mock)

        # -- Act -------------------------------------------------------------

        with pytest.raises(uut.SerializeError):
            uut.serialize(obj)
  
    @patch('energytt_platform.bus.serialize.json_serializer')
    def test__deserialize__message_not_in_registry__should_raise(self, json_serializer_mock):

        # -- Arrange ---------------------------------------------------------

        registry_mock = MagicMock()
        registry_mock.__contains__.return_value = False

        wrapped_msg_mock = MagicMock(type=123)

        json_serializer_mock.deserialize.return_value = wrapped_msg_mock

        uut = MessageSerializer(registry=registry_mock)

        # -- Act -------------------------------------------------------------

        with pytest.raises(uut.DeserializeError):
            uut.deserialize(b'does not matter')



"""
conftest.py according to pytest docs:
https://docs.pytest.org/en/2.7.3/plugins.html?highlight=re#conftest-py-plugins
"""
import pytest
from unittest.mock import patch
from testcontainers.postgres import PostgresContainer

from energytt_platform.sql import POSTGRES_VERSION, SqlEngine

from .db import db as _db


@pytest.fixture(scope='function')
def psql_uri():
    """
    TODO
    """
    image = f'postgres:{POSTGRES_VERSION}'

    with PostgresContainer(image) as psql:
        yield psql.get_connection_url()


@pytest.fixture(scope='function')
def db(psql_uri: str):
    """
    TODO
    """
    with patch('tests.integrationtest.sql.db.db.uri', new=psql_uri):
        yield _db


@pytest.fixture(scope='function')
def session(db: SqlEngine):
    """
    TODO
    """
    db.apply_schema()

    with db.make_session() as session:
        yield session



"""
Example setup.py from https://github.com/activescott/python-package-example/blob/master/package-project/src/setup.py.
"""
# Always prefer setuptools over distutils
import os
import setuptools
from distutils.command.sdist import sdist


class sdist_hg(sdist):
    """Add git short commit hash to version.
    Based on https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/specification.html
    """

    user_options = sdist.user_options + [
        ('dev', None, 'Add a dev marker'),
        ('build=', None, 'Build number'),
    ]

    def initialize_options(self):
        sdist.initialize_options(self)
        self.dev = None
        self.build = None

    def run(self):
        if self.build:
            if self.build.startswith('+'):
                prefix = ''
            else:
                prefix = '.'
            self.distribution.metadata.version += f'{prefix}{self.build}'
            print(self.distribution.metadata.version)
        sdist.run(self)

    def get_tip_revision(self):
        import git
        repo = git.Repo()
        sha = repo.head.commit.hexsha
        short_sha = repo.git.rev_parse(sha, short=True)
        return short_sha

    def get_timestamp(self):
        from datetime import datetime
        now = datetime.now()
        stamp = now.strftime("%Y%m%d%H%M%S")
        return stamp


here = os.path.abspath(os.path.dirname(__file__))


def __read_meta(fn):
    with open(os.path.join(here, 'meta', fn)) as f:
        return f.read().strip()


name = __read_meta('PACKAGE_NAME')
version = __read_meta('PACKAGE_VERSION')
python_requires = __read_meta('PYTHON_VERSION')


setuptools.setup(
    name=name,
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,
    description="Test project",
    author="TODO",
    author_email="jakob@jakobk.dk",
    # Choose your license
    license="Apache Software License 2.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: %s" % python_requires,
    ],
    # What does your project relate to?
    keywords=[],
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=setuptools.find_packages(),  # Required
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # TODO SPECIFIC VERSION OF REQUIREMENTS!
    # TODO Read from requirements.txt instead
    install_requires=[
        'wrapt',
        'pyjwt[crypto]',
        'kafka-python',
        'Flask',
        'serpyco',
        'rapidjson',
        'psycopg2',
        'sqlalchemy',
        'alembic',
        'requests',
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": [],
        "build": [],
        "test": [],
    },
    cmdclass={"sdist": sdist_hg},

    python_requires=">=3.8",

    include_package_data=True,
    package_data={'': [
        'meta/*',
        # 'static/*',
        # 'static/model-template/*',
    ]},
)



from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.common import DateTimeRange
from energytt_platform.models.measurements import Measurement


@dataclass
class MeasurementUpdate(Message):
    """
    A new Measurement has been added to the system.
    """
    measurement: Measurement


@dataclass
class MeasurementRemoved(Message):
    """
    TODO
    """
    id: str


@dataclass
class ImportMeasurements(Message):
    """
    TODO

    begin.from_ is INCLUDED.
    begin.to_ is EXCLUDED.
    """
    gsrn: str
    begin: DateTimeRange



import re
from abc import abstractmethod
from typing import Dict, Optional
from functools import cached_property

from energytt_platform.tokens import TokenEncoder
from energytt_platform.models.auth import InternalToken
from energytt_platform.auth import TOKEN_HEADER_NAME, TOKEN_COOKIE_NAME

from .responses import Unauthorized


class Context(object):
    """
    Context for a single incoming HTTP request.

    An instance is create for each HTTP request, and dies when the request
    if out of scope again.
    """

    # Regex pattern for matching bearer token (in HTTP header)
    TOKEN_PATTERN = re.compile(r'^Bearer:\s*(.+)$', re.IGNORECASE)

    def __init__(self, token_encoder: TokenEncoder[InternalToken]):
        """
        :param token_encoder: Internal token encoder
        """
        self.token_encoder = token_encoder

    @property
    @abstractmethod
    def headers(self) -> Dict[str, str]:
        """
        :returns: HTTP request headers
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def cookies(self) -> Dict[str, str]:
        """
        :returns: HTTP request cookies
        """
        raise NotImplementedError

    # -- Tokens --------------------------------------------------------------

    @property
    def opaque_token(self) -> Optional[str]:
        """
        Returns value for the opaque token provided by the client in a cookie.

        :returns: Opaque token or None
        """
        return self.cookies.get(TOKEN_COOKIE_NAME)

    @property
    def internal_token_encoded(self) -> Optional[str]:
        """
        Returns value for the raw, encoded internal token. The opaque token
        (provided by the client in a cookie) is translated to an internal
        token by the API gateway and passed on as a header.

        :returns: Raw, encoded internal token
        """
        if TOKEN_HEADER_NAME in self.headers:
            matches = self.TOKEN_PATTERN \
                .findall(self.headers[TOKEN_HEADER_NAME])

            if matches:
                return matches[0]

    @cached_property
    def token(self) -> Optional[InternalToken]:
        """
        Parses token into an OpaqueToken.
        """
        if self.internal_token_encoded is None:
            return None

        try:
            internal_token = self.token_encoder.decode(
                self.internal_token_encoded)
        except self.token_encoder.DecodeError:
            # TODO Raise exception if in debug mode?
            return None

        if internal_token.is_expired:
            # TODO Raise exception if in debug mode?
            return None

        return internal_token

    @property
    def is_authorized(self) -> bool:
        """
        Check whether or not the client provided a valid token.
        """
        return self.token is not None

    def has_scope(self, scope: str) -> bool:
        """
        TODO
        """
        if self.token:
            return scope in self.token.scope
        return False

    def get_token(self, required=True) -> Optional[InternalToken]:
        """
        TODO
        """
        if self.token:
            return self.token
        elif required:
            raise Unauthorized('')  # TODO Error message

    def get_subject(self, required=True) -> Optional[str]:
        """
        TODO
        """
        if self.token:
            return self.token.subject
        elif required:
            raise Unauthorized('')  # TODO Error message



from dataclasses import dataclass

from energytt_platform.bus import Message
from energytt_platform.models.tech import Technology, TechnologyCodes


@dataclass
class TechnologyUpdate(Message):
    """
    A Technology has been added or updated.
    """
    technology: Technology


@dataclass
class TechnologyRemoved(Message):
    """
    TODO
    """
    codes: TechnologyCodes



from abc import abstractmethod
from typing import Optional, Any, Type
from inspect import getfullargspec
from functools import cached_property


class Endpoint(object):
    """
    Represents a single HTTP API endpoint.
    """

    # Request and response schemas (dataclasses or None)
    Request = None
    Response = None

    @abstractmethod
    def handle_request(self, **kwargs) -> Optional[Any]:
        """
        Handle a HTTP request.
        """
        raise NotImplementedError

    @property
    def request_schema(self) -> Optional[Type[Any]]:
        """
        TODO
        """
        return self.Request

    @property
    def response_schema(self) -> Optional[Type[Any]]:
        """
        TODO
        """
        return self.Response

    @cached_property
    def requires_context(self) -> bool:
        """
        Returns True if handle_request() requires a Context passed as argument.
        """
        return 'context' in getfullargspec(self.handle_request)[0]

    @cached_property
    def should_parse_request_data(self) -> bool:
        """
        Returns True if handle_request() requires an instance of self.Request
        passed as argument.
        """
        return (self.request_schema is not None
                and 'request' in getfullargspec(self.handle_request)[0])



from energytt_platform.bus import MessageBroker, Message


class TestMessageBroker:
    """
    TODO
    """

    # -- poll() --------------------------------------------------------------

    def test__poll__publish_messages_and_subscribe_to_all_topics__should_receive_messages_from_all_topics(
            self,
            broker: MessageBroker,
            msg1: Message,
            msg2: Message,
            msg3: Message,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        broker.publish(topic='TOPIC1', msg=msg1)
        broker.publish(topic='TOPIC2', msg=msg2)
        broker.publish(topic='TOPIC2', msg=msg3)

        broker.subscribe(['TOPIC1', 'TOPIC2'])

        received_messages = broker.poll(timeout=5)

        # -- Assert ----------------------------------------------------------

        assert received_messages == {
            'TOPIC1': [msg1],
            'TOPIC2': [msg2, msg3],
        }

    def test__poll__publish_messages_and_subscribe_to_some_topics__should_receive_messages_from_subscribed_topics(
            self,
            broker: MessageBroker,
            msg1: Message,
            msg2: Message,
            msg3: Message,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        broker.publish(topic='TOPIC1', msg=msg1)
        broker.publish(topic='TOPIC2', msg=msg2)
        broker.publish(topic='TOPIC3', msg=msg3)

        broker.subscribe(['TOPIC1'])

        received_messages = broker.poll(timeout=5)

        # -- Assert ----------------------------------------------------------

        assert received_messages == {
            'TOPIC1': [msg1],
        }

    # -- poll_list() ---------------------------------------------------------

    def test__poll_list__publish_messages_and_subscribe_to_all_topics__should_receive_messages_from_all_topics(
            self,
            broker: MessageBroker,
            msg1: Message,
            msg2: Message,
            msg3: Message,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        broker.publish(topic='TOPIC1', msg=msg1)
        broker.publish(topic='TOPIC2', msg=msg2)
        broker.publish(topic='TOPIC3', msg=msg3)

        broker.subscribe(['TOPIC1', 'TOPIC2', 'TOPIC3'])

        received_messages = broker.poll_list(timeout=5)

        # -- Assert ----------------------------------------------------------

        assert len(received_messages) == 3
        assert all(msg in received_messages for msg in [msg1, msg2, msg3])

    def test__poll_list__publish_messages_and_subscribe_to_some_topics__should_receive_messages_from_all_topics(
            self,
            broker: MessageBroker,
            msg1: Message,
            msg2: Message,
            msg3: Message,
    ):
        """
        TODO
        """

        # -- Act -------------------------------------------------------------

        broker.publish(topic='TOPIC1', msg=msg1)
        broker.publish(topic='TOPIC2', msg=msg2)
        broker.publish(topic='TOPIC3', msg=msg3)

        broker.subscribe(['TOPIC1'])

        received_messages = broker.poll_list(timeout=5)

        # -- Assert ----------------------------------------------------------

        assert received_messages == [msg1]



from typing import List
from uuid import uuid4

import pytest
from flask.testing import FlaskClient

from energytt_platform.api import (
    Application,
    EndpointGuard,
    TokenGuard,
    ScopedGuard,
)

from .endpoints import EmptyEndpoint


class TestGuards:
    """
    TODO
    """

    @pytest.mark.parametrize('guard', [
        TokenGuard(),
        ScopedGuard('scope1'),
    ])
    def test__provide_no_token__should_return_status_401(
            self,
            guard: EndpointGuard,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EmptyEndpoint(),
            guards=[guard],
        )

        # -- Act -------------------------------------------------------------

        r = client.post('/something')

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 401

    @pytest.mark.parametrize('guard', [
        TokenGuard(),
        ScopedGuard('scope1'),
    ])
    def test__provide_invalid_token__should_return_status_401(
            self,
            guard: EndpointGuard,
            app: Application,
            client: FlaskClient,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EmptyEndpoint(),
            guards=[guard],
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: NOT-A-VALID-TOKEN'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 401

    @pytest.mark.parametrize('guard', [
        TokenGuard(),
        ScopedGuard('scope1'),
    ])
    def test__provide_valid_token__should_return_status_200(
            self,
            guard: EndpointGuard,
            app: Application,
            client: FlaskClient,
            valid_token_encoded: str,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EmptyEndpoint(),
            guards=[guard],
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: {valid_token_encoded}'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200

    def test__token_missing_required_scope__should_return_status_401(
            self,
            app: Application,
            client: FlaskClient,
            valid_token_encoded: str,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        required_scope = str(uuid4())  # Something random

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EmptyEndpoint(),
            guards=[ScopedGuard(required_scope)],
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: {valid_token_encoded}'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 401

    @pytest.mark.parametrize('guards', [
        [ScopedGuard('scope1')],
        [ScopedGuard('scope2')],
        [ScopedGuard('scope1', 'scope2')],
        [TokenGuard(), ScopedGuard('scope1')],
        [TokenGuard(), ScopedGuard('scope1', 'scope2')],
        [TokenGuard(), ScopedGuard('scope1'), ScopedGuard('scope2')],
    ])
    def test__token_has_required_scope__should_return_status_200(
            self,
            guards: List[EndpointGuard],
            app: Application,
            client: FlaskClient,
            valid_token_encoded: str,
    ):
        """
        TODO
        """

        # -- Arrange ---------------------------------------------------------

        app.add_endpoint(
            method='POST',
            path='/something',
            endpoint=EmptyEndpoint(),
            guards=guards,
        )

        # -- Act -------------------------------------------------------------

        r = client.post(
            path='/something',
            headers={'Authorization': f'Bearer: {valid_token_encoded}'},
        )

        # -- Assert ----------------------------------------------------------

        assert r.status_code == 200



from typing import Dict, Type

from .broker import Message, TMessageHandler


class MessageDispatcher(Dict[Type[Message], TMessageHandler]):
    """
    A message handler that dispatches incoming messages to the appropriate
    handler. Each message type can have a single handler associated.
    """
    def __call__(self, msg: Message):
        message_type = type(msg)
        if message_type in self:
            print('DISPATCH: %s' % msg)
            handler = self[message_type]
            handler(msg)
        else:
            print('IGNORING: %s' % message_type)



from dataclasses import dataclass

from energytt_platform.bus import Message


@dataclass
class UserOnboarded(Message):
    """
    A new user has been onboarded to the system.
    """
    subject: str



from typing import List
from datetime import datetime
from dataclasses import dataclass

from energytt_platform.serialize import Serializable
from energytt_platform.models.tech import TechnologyType

from .emissions import EmissionValues


# -- Residual Mix emissions --------------------------------------------------


@dataclass
class ResidualMixEmissions(Serializable):
    """
    Describes emissions in the general mix within a single
    sector (pricing area), and in a specific time-frame.
    """

    # Sector where energy is consumed
    sector: str

    # The time-frame where the distribution is valid.
    # TODO Is 'end' included or excluded in the period?
    begin: datetime
    end: datetime

    # Relative emissions (gram/Wh)
    emissions: EmissionValues


# -- Residual Mix technologies -----------------------------------------------


@dataclass
class ResidualMixTechnology(Serializable):
    """
    Describes a part of the residual mix consumption, specifically
    how large a percentage of the total energy consumed originates
    from a single type of technology.
    """
    technology: TechnologyType
    percent: float
    sector: str  # Sector where energy is produced


@dataclass
class ResidualMixTechnologyDistribution(Serializable):
    """
    Describes the distribution of technologies used to produce energy
    in the residual mix within a single sector (pricing area),
    and in a specific time-frame.
    """

    # Sector where energy is consumed
    sector: str

    # The time-frame where the distribution is valid.
    # TODO Is 'end' included or excluded in the period?
    begin: datetime
    end: datetime

    # Distribution of produced energy on technologies.
    # The list is unique on (technology, sector) (composite key).
    # The sum of ResidualMixTechnology.percent MUST be exactly 1.0.
    technologies: List[ResidualMixTechnology]



from datetime import datetime
from dataclasses import dataclass

from energytt_platform.serialize import Serializable

from .common import EnergyDirection


MeasurementType = EnergyDirection


@dataclass
class Measurement(Serializable):
    id: str
    gsrn: str
    amount: int
    begin: datetime
    end: datetime
